package org.jastadd.ros2rag.compiler;

import beaver.Parser;
import org.jastadd.ros2rag.ast.*;
import org.jastadd.ros2rag.compiler.options.CommandLine;
import org.jastadd.ros2rag.compiler.options.CommandLine.CommandLineException;
import org.jastadd.ros2rag.compiler.options.Option;
import org.jastadd.ros2rag.compiler.options.StringOption;
import org.jastadd.ros2rag.parser.RelAstParser;
import org.jastadd.ros2rag.scanner.RelAstScanner;

import java.io.*;
import java.util.ArrayList;
import java.util.List;

public class Compiler {

  private StringOption optionOutputDir;

  private ArrayList<Option<?>> options;
  private CommandLine commandLine;

  public Compiler(String[] args) throws CommandLineException {
    options = new ArrayList<>();
    addOptions();

    commandLine = new CommandLine(options);
    commandLine.parse(args);

    String outputDir = "gen"; // should not be used, but otherwise there is a compiler warning
    if (optionOutputDir.isSet()) {
      outputDir = optionOutputDir.getValue();
    } else {
      System.out.println("No output grammar name is set. Assuming output dir '" + outputDir + "'.");
    }

    printMessage("Running ROS2RAG Preprocessor");

    if (commandLine.getArguments().size() < 1) {
      error("specify at least one input file");
    }

    List<String> filenames = commandLine.getArguments();
    Program p = parseProgram(filenames);

    writeToFile(outputDir + "/Grammar.ast", p.generateAbstractGrammar());
    writeToFile(outputDir + "/ROS2RAG.jadd", p.generateAspect());

  }


  public static void main(String[] args) {
    try {
      new Compiler(args);
    } catch (CommandLineException e) {
      System.err.println(e.getMessage());
      System.exit(1);
    }
  }

  private void printMessage(String message) {
    System.out.println(message);
  }

  private void writeToFile(String filename, String str) {
    try {
      PrintWriter writer = new PrintWriter(filename);
      writer.print(str);
      writer.close();
    } catch (Exception e) {
      e.printStackTrace();
      System.exit(1);
    }
  }

  private void addOptions() {
    optionOutputDir = addOption(new StringOption("outputDir", "target directory for the generated files."));
  }

  private <OptionType extends Option<?>> OptionType addOption(OptionType option) {
    options.add(option);
    return option;
  }

  private Program parseProgram(List<String> fileNames) {

    Program program = new Program();

    for (String fileName : fileNames) {
      FileReader reader = null;
      try {
        reader = new FileReader(fileName);
      } catch (FileNotFoundException e) {
        System.err.println(e.getMessage());
        System.exit(1);
      }

      parse(program, reader, fileName);
    }
    program.treeResolveAll();
    return program;
  }

  private void parse(Program program, Reader reader, String file) {
    RelAstScanner scanner = new RelAstScanner(reader);
    RelAstParser parser = new RelAstParser();

    try {
      GrammarFile newFile = (GrammarFile) parser.parse(scanner);
      program.addGrammarFile(newFile);
    } catch (IOException e) {
      error(e.getMessage());
    } catch (Parser.Exception e) {
      System.err.println("Parse error in file " + file);
      System.err.println(e.getMessage());
      System.exit(1);
    }
  }

  protected void error(String message) {
    System.err.println("Error: " + message);
    System.err.println();
    System.err.println("Usage: java -jar ros2rag.jar [--option1] [--option2=value] ...  <filename1> <filename2> ... ");
    System.err.println("Options:");
    System.err.print(commandLine.printOptionHelp());
    System.exit(1);
  }
}

